# coding=utf-8
import logging

import six

from .luqum.tree import OrOperation, Word
from .luqum.utils import LuceneTreeVisitorV2, UnknownOperationResolver
from .luqum.parser import parser, lexer
import pyparsing as pp


def process_vals(t, l):
    res = []
    for val in t:
        res.append(l(val))

    return res


def convert_number(t):
    def int_or_float(val):
        if '.' in val:
            return float(val)

        return int(val)

    return process_vals(t, int_or_float)


def parse_expr(expr_text):
    operator = pp.Regex(">=|<=|!=|>|<|=")('operator')

    point = pp.Literal('.')
    e = pp.CaselessLiteral('E')
    plus_or_minus = pp.Literal('+') | pp.Literal('-')
    number = pp.Word(pp.nums)
    integer = pp.Combine(pp.Optional(plus_or_minus) + number)
    float_number = pp.Combine(integer + pp.Optional(point + pp.Optional(number)) + pp.Optional(e + integer))

    float_number.setParseAction(convert_number)

    quoted_string = pp.QuotedString('"')
    identifier = pp.Word(pp.alphanums + "_" + "." + "-")
    text_identifier = (quoted_string | identifier)
    comparison_term = float_number | text_identifier
    values = pp.OneOrMore(comparison_term)('values')
    range_op = pp.Suppress(pp.CaselessLiteral('TO'))
    range_operator = (pp.Suppress('[') + comparison_term + range_op + comparison_term + pp.Suppress(']'))('range')
    range_inclusive = (pp.Suppress('{') + comparison_term + range_op + comparison_term + pp.Suppress('}'))('range_inclusive')

    condition_group = pp.Group((range_inclusive | range_operator | values | (operator + values)))('condition_group')
    conditions = pp.Group(condition_group)('conditions')

    expr = pp.infixNotation(conditions, [
        ("AND", 2, pp.opAssoc.LEFT,),
        ("OR", 2, pp.opAssoc.LEFT,),
    ])

    return expr.parseString(expr_text)


def __conditions_expr_to_sql_where(field_name, item):
    return '({sql})'.format(sql=expr_to_sql_where(field_name, item))


def quotes_if_needed(val):
    return '"%s"' % val if isinstance(val, six.string_types) else val


def __condition_group_expr_to_sql_where(field_name, item):
    if 'values' in item:
        values = item.get('values')
        operator = item.get('operator', '=')
        vals = []
        for val in values:
            vals.append('{field_name}{operator}{value}'.format(
                field_name=field_name, operator=operator, value=quotes_if_needed(val)))

        return 'OR'.join(vals)

    if 'range' in item:
        return ('({field_name}>={value1})AND({field_name}<={value2})'.format(
            field_name=field_name, value1=item[0], value2=item[1]))

    if 'range_inclusive' in item:
        return ('({field_name}>{value1})AND({field_name}<{value2})'.format(
            field_name=field_name, value1=item[0], value2=item[1]))


def expr_to_sql_where(field_name, expr):
    sql = ''
    for item in expr:
        if item.getName() == 'conditions':
            sql += __conditions_expr_to_sql_where(field_name, item)
        elif item.getName() == 'condition_group':
            sql += __condition_group_expr_to_sql_where(field_name, item)

    return sql


class SQLQueryBuilder(LuceneTreeVisitorV2):
    def __init__(self, sql_helper):
        self.__where = {}
        self.__limit = []
        self.__vars = {}
        self.__sql_helper = sql_helper

    def __handle_internal_function(self, node):
        if node.name.lower() == 'sample':
            sample = 1.0 - float(node.expr.value)

            random_function = self.__sql_helper.random_function_name()

            self.__where.setdefault(1, []).append(('AND', '({random_function}>{sample:.4g})'.format(random_function=random_function, sample=sample)))
            return True

        if node.name.lower() == 'seed':
            self.__vars['seed'] = int(str(node.expr))
            return True

        if node.name.lower() == 'version':
            self.__vars['version'] = str(node.expr)
            return True

        if node.name.lower() == 'limit':
            self.__vars['limit'] = int(str(node.expr))
            return True

    def visit_prohibit(self, node, parents, context):
        self.__add_where(context.get('field_name'), str(node), context.get('op'))

    def visit_phrase(self, node, parents, context):
        self.__add_where(context.get('field_name'), str(node)[1:], context.get('op'))

    def visit_range(self, node, parents, context):
        self.__add_where(context.get('field_name'), str(node), context.get('op'))

    def visit_plus(self, node, parents, context):
        self.__add_where(context.get('field_name'), str(node)[1:], context.get('op'))

    def visit_field_group(self, node, parents, context):
        self.visit(node.expr, parents + [node], context)

    def __add_where(self, field_name, expr, op):
        where = expr_to_sql_where(field_name, parse_expr(expr))
        self.__where.setdefault(0, []).append((op, where))

    def visit_word(self, node, parents, context):
        self.__add_where(context.get('field_name'), str(node), context.get('op'))

    def visit_search_field(self, node, parents, context):
        if self.__handle_internal_function(node):
            return

        context = context or {}

        for child in node.children:
            if isinstance(child, Word):
                self.__add_where(node.name, str(node.expr), context.get('op'))
            else:
                self._sub_visit(node.expr, None, context={'field_name': node.name})

    @classmethod
    def __combine_where(cls, op, sql_builder):
        combine_where = {}
        for bucket, wheres in sorted(sql_builder.where.items()):
            for where in wheres:
                operator, expr = where
                operator = operator or op

                combine_where.setdefault(bucket, []).append((operator, expr))

        return combine_where

    def _sub_visit(self, child, op, context=None):
        sql_builder = SQLQueryBuilder(self.__sql_helper)
        sql_builder.visit(child, context=context)

        self.__vars.update(sql_builder.vars)

        combine_where = self.__combine_where(op, sql_builder)

        for bucket, wheres in combine_where.items():
            op = None
            sql = ''
            for where in wheres:
                operator, expr = where
                if op is not None:
                    sql += op

                sql += expr
                op = operator

            if len(wheres) > 1:
                sql = self.wrap_in_parentheses(sql)

            self.__where.setdefault(bucket, []).append((None, sql))

    def __visit_binary_operation(self, node, parents, context, op):
        context = context or {}
        context['op'] = op
        for child in node.children:
            self.visit(child, context=context)

    def visit_and_operation(self, node, parents, context):
        self.__visit_binary_operation(node, parents, context, 'AND')

    def visit_or_operation(self, node, parents, context):
        self.__visit_binary_operation(node, parents, context, 'OR')

    @classmethod
    def wrap_in_parentheses(cls, text):
        return '(%s)' % text

    def build_vars(self):
        if not self.__vars:
            return None

        return self.__vars

    def __build_where(self):
        sql = ''
        for bucket, wheres in sorted(self.__where.items()):
            last_operator = None

            for where in wheres:
                operator, expr = where

                if last_operator:
                    sql += last_operator
                elif sql:
                    sql += operator

                sql += expr

                last_operator = operator

        return sql

    def build_where(self):
        return self.__build_where() or None  # Make sure to return None and not empty string

    @property
    def where(self):
        return self.__where

    @property
    def vars(self):
        return self.__vars


def parse_query(query):
    return parser().parse(query, lexer=lexer())


def tree_to_sql_parts(tree, sql_helper):
    sql_builder = SQLQueryBuilder(sql_helper)
    resolver = UnknownOperationResolver(OrOperation)
    sql_builder.visit(resolver(tree))

    return sql_builder.build_vars(), sql_builder.build_where()


if __name__ == '__main__':  # pragma: no cover
    parse_query('version:a196d73071e57874768e20267a1bec4ed0a8c2b7')



# coding=utf-8
import datetime
import logging

import six
from ..bigquery_mixin import BigQueryMixin
from .base_metadata_db import BaseMetadataDB, MetadataTypeNotSupported
from flatten_json import unflatten


# noinspection SqlNoDataSourceInspection
class BigQueryMetadataDB(BaseMetadataDB, BigQueryMixin):
    STAGING_TABLE_NAME = 'staging'
    METADATA_TABLE_NAME = 'metadata'
    INDEX_TABLE_NAME = 'index'
    STAGING_INDEX_TABLE_NAME = 'staging_index'

    def __init__(self, connection, version=None, delete_temp_on_commit=True):
        self.__version = version
        self.__staging_table = None
        self.__metadata_table = None
        self.__prev_table_info = None
        self.__delete_temp_on_commit = delete_temp_on_commit
        super(BigQueryMetadataDB, self).__init__(connection)

    def __get_staging_table_name(self):
        version = self.__version or 0
        return '%s_%s' % (self.STAGING_TABLE_NAME, version)

    def __get_staging_index_table_name(self):
        version = self.__version or 0
        return '%s_%s' % (self.STAGING_INDEX_TABLE_NAME, version)

    @classmethod
    def _default_table_schema(cls):
        from google.cloud import bigquery

        schema = (
            bigquery.SchemaField('_sha', 'STRING'),
            bigquery.SchemaField('_commit_sha', 'STRING'),
            bigquery.SchemaField('_ts', 'TIMESTAMP'),
            bigquery.SchemaField('_hash', 'STRING'),
            bigquery.SchemaField('url', 'STRING'),
        )

        return schema

    def _create_table(self):
        schema = self._default_table_schema()

        self.__metadata_table = self._create_specific_table(self.METADATA_TABLE_NAME, schema)
        self.__staging_table = self._create_specific_table(self.__get_staging_table_name(), self.__metadata_table.schema)

    def __has_column(self, column_name):
        for field in self.__metadata_table.schema:
            if column_name == field.name:
                return True

        return False

    @classmethod
    def __value_to_sql_type(cls, column_value):
        if isinstance(column_value, six.string_types):
            return 'STRING'

        if isinstance(column_value, six.integer_types):
            return 'INTEGER'

        if isinstance(column_value, float):
            return 'FLOAT'

        raise MetadataTypeNotSupported('UNKNOWN TYPE %s' % type(column_value))

    @classmethod
    def _patch_table_with_new_schema(cls, table, schema):
        table.patch(schema=schema)

    def _add_missing_columns(self, data_object):
        from google.cloud import bigquery

        new_columns = []
        for column_name, column_value in data_object.items():
            column_name = self.common_name_to_bq_field_name(column_name)
            if self.__has_column(column_name):
                continue

            new_columns.append(bigquery.SchemaField(column_name, self.__value_to_sql_type(column_value)))

        if len(new_columns) > 0:
            schema = self.__staging_table.schema[:]
            new_columns = sorted(new_columns, key=lambda field: field.name)
            schema.extend(new_columns)
            self._patch_table_with_new_schema(self.__staging_table, schema)
            self._patch_table_with_new_schema(self.__metadata_table, schema)

    def _add_data(self, flatten_data_list):
        rows = []

        now = datetime.datetime.utcnow()

        for flatten_data_list in flatten_data_list:
            row = []
            for field in self.__staging_table.schema:
                common_field_name = self.bq_field_name_to_common_name(field.name)
                if common_field_name == '_ts':
                    row.append(now)
                else:
                    row.append(flatten_data_list.get(common_field_name))

            rows.append(row)

        self.handle_bq_errors(self.__staging_table.insert_data(rows))

    def __truncate_staging(self):
        if not self.__delete_temp_on_commit:
            logging.debug('meta: delete_temp_on_commit')
            return

        logging.info('truncate metadata staging')
        staging_table = self._get_table(self.__get_staging_table_name())
        staging_table.delete()

    def end_commit(self):
        logging.debug('bq end commit meta')

        self.__truncate_staging()

    def begin_commit(self, commit_sha, tree_id):
        logging.debug('bq begin commit meta %s %s', commit_sha, tree_id)

        from google.cloud import bigquery

        staging_table = self._get_table(self.__get_staging_table_name())
        index_staging_table = self._get_table(self.__get_staging_index_table_name())
        metadata_table = self._get_table(self.METADATA_TABLE_NAME)

        with self._connection.get_cursor() as bq_dataset:
            src_query = """
                #standardSQL
                SELECT @commit_sha as _commit_sha, * 
                # Find the latest metadata given in the staging meta
                FROM (
                  SELECT * EXCEPT(row_number)
                  FROM (
                    # Find the latest metadata given in the staging meta
                    SELECT * EXCEPT(_hash, _commit_sha) , ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number 
                    FROM {dataset_name}.{staging_table_name} 
                  )
                  WHERE row_number = 1
                )  
            """.format(
                dataset_name=bq_dataset.name,
                staging_table_name=staging_table.name,
                staging_index_table_name=index_staging_table.name)

            src_query_parameters = (
                bigquery.ScalarQueryParameter('commit_sha', 'STRING', commit_sha),
            )

            self._copy_table_data(src_query, src_query_parameters, metadata_table)

    def _query_head_data(self, sha_list):
        from google.cloud import bigquery

        bq_client = self._connection

        with bq_client.get_cursor() as bq_dataset:
            query = """
            #standardSQL
            SELECT * EXCEPT(_max_sha, _max_ts, _ts) FROM (SELECT *
              FROM `{dataset_name}.{staging_table_name}` WHERE _sha IN UNNEST(@sha_list)
              UNION ALL
                SELECT * FROM `{dataset_name}.{metadata_table_name}` WHERE _sha IN UNNEST(@sha_list)) as metadata_staging_combine
              INNER JOIN (
                SELECT  _sha as _max_sha, MAX(_ts) as _max_ts FROM `{dataset_name}.{staging_table_name}` WHERE _sha IN UNNEST(@sha_list) GROUP BY _sha
                UNION ALL
                  SELECT  _sha as _max_sha, MAX(_ts) as _max_ts FROM `{dataset_name}.{metadata_table_name}` WHERE _sha IN UNNEST(@sha_list) GROUP BY _sha
              ) _max_metadata
            ON
              metadata_staging_combine._sha = _max_sha
            WHERE
              metadata_staging_combine._ts = _max_ts;
            """.format(
                dataset_name=bq_dataset.name,
                staging_table_name=self._get_table_name(bq_client.table_prefix, self.__get_staging_table_name()),
                metadata_table_name=self._get_table_name(bq_client.table_prefix, self.METADATA_TABLE_NAME))

            query_parameters = (
                bigquery.ArrayQueryParameter('sha_list', 'STRING', sha_list),
            )

            items_iter, _ = self._query_sync(query, query_parameters, process_row=self.build_dict)

            return items_iter

    def _query(self, sql_vars, select_fields, where, max_results=None, start_index=None):
        import google.api.core.exceptions

        bq_client = self._connection

        metadata_table_name = self._get_table_name(bq_client.table_prefix, self.METADATA_TABLE_NAME)
        staging_table_name = self._get_table_name(bq_client.table_prefix, self.__get_staging_table_name())
        staging_index_table_name = self._get_table_name(bq_client.table_prefix, self.__get_staging_index_table_name())
        index_table_name = self._get_table_name(bq_client.table_prefix, self.INDEX_TABLE_NAME)

        limit = sql_vars.get('limit')
        limit = 'LIMIT %s' % limit if limit is not None else ''

        def query_staging():
            query = """
                #standardSQL
                SELECT url as _url, * EXCEPT(_ts, url)
                FROM ( # Bring all the items from staging index (except the metadata files)
                  SELECT *
                  FROM (
                    SELECT name as _sha
                    FROM (
                      SELECT *, ROW_NUMBER() OVER (PARTITION BY name) row_number
                      FROM {dataset_name}.{staging_index_table_name}
                      WHERE ENDS_WITH(name, '.metadata') = FALSE 
                     )
                    WHERE row_number = 1
                  ) staging_index
                  FULL JOIN (
                    SELECT _sha
                    FROM (
                      SELECT _sha, ROW_NUMBER() OVER (PARTITION BY _sha) row_number
                      FROM {dataset_name}.{staging_table_name}
                    )
                    WHERE row_number = 1    
                  ) staging_meta
                  USING(_sha)  
                ) AS staging_items
                LEFT JOIN ( 
                  # join this with latest (max _ts) metadata
                  SELECT * EXCEPT(_commit_sha)
                  FROM (
                    SELECT * EXCEPT(_hash, url)
                    FROM {dataset_name}.{staging_table_name}
                    UNION ALL
                    SELECT * EXCEPT(_hash, url)
                    FROM {dataset_name}.{metadata_table_name}
                    WHERE _ts = (
                        SELECT MAX(_ts)
                        FROM (
                          SELECT *
                          FROM {dataset_name}.{staging_table_name}
                          UNION ALL
                          SELECT *
                          FROM {dataset_name}.{metadata_table_name}
                        ) 
                        GROUP BY _sha
                        LIMIT 1
                      )
                  ) meta
                  RIGHT JOIN ( 
                    # This join will give us the latest URL (url) and hash (_hash) of the data (by name)
                    SELECT name as _sha, sha as _hash, url
                    FROM ( # This will remove any duplicates we might have
                      SELECT * EXCEPT(row_number)
                      FROM (
                        SELECT *, ROW_NUMBER() OVER (PARTITION BY name) row_number
                        FROM (
                          SELECT *
                          FROM {dataset_name}.{staging_index_table_name}
                          UNION ALL
                          SELECT *
                          FROM {dataset_name}.{index_table_name}
                          ORDER by ts DESC
                        ) 
                      )
                      WHERE row_number = 1
                    )
                  ) index_with_hash
                  USING(_sha)
                ) AS meta
                USING(_sha)
                WHERE({where})
                ORDER BY _ts DESC, _sha
                {limit}
             """.format(
                dataset_name=bq_dataset.name,
                index_table_name=index_table_name,
                staging_table_name=staging_table_name,
                metadata_table_name=metadata_table_name,
                staging_index_table_name=staging_index_table_name,
                where=where or 'True',
                select=','.join(select_fields),
                limit=limit)

            return query

        def query_all_meta_data_without_staging(version):
            query = """
                #standardSQL
                SELECT *
                FROM (
                    SELECT * EXCEPT(row_number, _hash, url, _ts)
                    FROM (
                      SELECT *, ROW_NUMBER() OVER (PARTITION BY _hash ORDER BY _ts DESC) row_number 
                      FROM (
                        SELECT * FROM {dataset_name}.{metadata_table_name} 
                        WHERE _ts <= ( 
                          SELECT MAX(_ts) 
                          FROM {dataset_name}.{metadata_table_name} 
                          WHERE _commit_sha = '{version}' 
                          LIMIT 1
                        ) 
                      )
                    )
                    WHERE row_number = 1 AND ({where})
                    ORDER BY FARM_FINGERPRINT(CONCAT(_sha, '$seed'))
                ) 
                LEFT JOIN (
                    # This join will give us the latest URL (url) and hash (_hash) of the data (by name)
                    SELECT name as _sha, sha as _hash, url as _url
                    FROM ( 
                    # This will remove any duplicates we might have
                      SELECT * EXCEPT(row_number)
                      FROM (
                        SELECT *, ROW_NUMBER() OVER (PARTITION BY name) row_number
                        FROM {dataset_name}.{index_table_name}
                        WHERE ts <= ( 
                          SELECT MAX(_ts) 
                          FROM {dataset_name}.{metadata_table_name}
                          WHERE _commit_sha = '{version}'
                          LIMIT 1
                        ) 
                        ORDER by ts DESC
                      )
                      WHERE row_number = 1
                    )
                )
                USING(_sha)
                {limit}              
             """.format(
                dataset_name=bq_dataset.name,
                version=version,
                metadata_table_name=metadata_table_name,
                index_table_name=index_table_name,
                where=where or 'True',
                select=','.join(select_fields),
                limit=limit)

            return query

        def query_everything():
            query = """
                 #Legacy SQL
                 select {select} EXCEPT(_ts)
                  from (
                    select * from `{dataset_name}.{metadata_table_name}`
                    UNION ALL
                    SELECT
                      * EXCEPT(row_number)
                    FROM (
                      SELECT
                        *,
                        ROW_NUMBER() OVER (PARTITION BY _sha ORDER BY _ts DESC) row_number
                      FROM
                        {dataset_name}.{staging_table_name})
                    WHERE
                      row_number = 1                              
                  ) 
                 where _ts IN (
                   SELECT MAX(_ts)
                   FROM {dataset_name}.{metadata_table_name}
                   GROUP BY _sha
                 ) AND ({where})
                 ORDER BY RAND($seed)
                 {limit}
             """.format(
                dataset_name=bq_dataset.name,
                metadata_table_name=metadata_table_name,
                staging_table_name=staging_table_name,
                where=where or 'True',
                select=','.join(select_fields),
                limit=limit)

            return query

        def _data_iter():
            for result in data_iter:
                for key, val in result.items():
                    if isinstance(val, six.string_types):
                        result[key] = val.encode('utf8')

                # if we don't have commit sha it means that this is a staging query
                result.setdefault('@commit_sha', 'staging')

                yield unflatten(result, separator='.')

        with bq_client.get_cursor() as bq_dataset:
            version_var = sql_vars.get('version')

            if version_var is None:
                query_sql = query_everything()
            elif version_var.lower() == 'staging':
                query_sql = query_staging()
            else:
                query_sql = query_all_meta_data_without_staging(version_var.lower())

            query_sql = self.fill_in_vars(query_sql, sql_vars)

            try:
                data_iter, total_rows = self._query_sync(
                    query_sql, process_row=self.build_dict, max_results=max_results, start_index=start_index)

                return _data_iter(), total_rows
            except google.api.core.exceptions.NotFound:
                return [], 0

    def get_all_data(self, sha):
        pass

    def get_data_for_commit(self, sha, commit_sha):
        pass



# coding=utf-8
from fudge import Fake

from tests.base import BaseTest
from mali_commands.legit.object_store.gcs.gcs_object_store import GCSDownload


class TestGCSDownload(BaseTest):

    def testGetRequest(self):
        object_name = 'object_name_' + self.some_random_shit()
        signed_url = 'signed_url_' + self.some_random_shit()

        signed_url_service = Fake()\
            .expects('get_signed_urls')\
            .with_args(['GET'], [object_name])\
            .returns({'GET': [signed_url]})

        gcs_downloader = GCSDownload(signed_url_service)
        request = gcs_downloader._get_request('', '', object_name)

        self.assertEqual(request.uri, signed_url)



# coding=utf-8
from .connection_mixin import ConnectionMixin


class DatastoreMixin(ConnectionMixin):
    def _build_key(self, datastore, name):
        org_name = self.__connection.data_volume_config.org
        parent_key = datastore.key('DataVolume', self.__connection.data_volume_config.name, namespace=org_name)

        return datastore.key(self._entity_kind, name, parent=parent_key, namespace=org_name)

    @property
    def _entity_kind(self):
        raise NotImplementedError(self._entity_kind)



#!/usr/bin/env python
# coding=utf-8
import click
import sys
import logging
from self_update.sdk_version import get_version, get_keywords


class Expando(object):
    pass


package = 'mali'


@click.group()
@click.option('--outputFormat', '-o', type=click.Choice(['tables', 'json']), default='tables', required=False)
@click.option('--configPrefix', '-cp', envvar='CONFIG_PREFIX', required=False)
@click.pass_context
def cli(ctx, outputformat, configprefix):
    from mali_commands.config import Config

    ctx.obj = Expando()

    from mali_commands.commons import handle_api

    config = Config(configprefix)

    ctx.obj.config_prefix = configprefix
    ctx.obj.handle_api = handle_api

    ctx.obj.api_host = config.api_host
    ctx.obj.host = config.host
    ctx.obj.client_id = config.client_id
    ctx.obj.refresh_token = config.refresh_token

    ctx.obj.auth0 = config.auth0
    ctx.obj.output_format = outputformat

    ctx.obj.refresh_token = config.refresh_token
    ctx.obj.id_token = config.id_token


@cli.command('version')
def version():
    click.echo('%s %s' % (package, __version__))

loaded = False
__version__ = get_version(package)
keywords = get_keywords(package) or []


# noinspection PyBroadException
def update_sdk(latest_version, user_path, throw_exception):
    from self_update.pip_util import pip_install, get_pip_server

    require_package = '%s==%s' % (package, latest_version)
    p, args = pip_install(get_pip_server(keywords), require_package, user_path)

    if p is None:
        return False

    try:
        std_output, std_err = p.communicate()
    except Exception:
        if throw_exception:
            raise

        logging.exception("%s failed", " ".join(args))
        return False

    rc = p.returncode

    if rc != 0:
        logging.error('%s failed to upgrade to latest version (%s)', package, latest_version)
        logging.error("failed to run %s (%s)\n%s\n%s", " ".join(args), rc, std_err, std_output)
        return False

    logging.info('%s updated to latest version (%s)', package, latest_version)

    return True


def self_update(throw_exception=False):
    from self_update.pip_util import get_latest_pip_version

    global __version__

    current_version = get_version(package)

    if current_version is None:
        __version__ = 'Please install this project with setup.py'
        return

    latest_version = get_latest_pip_version(package, keywords, throw_exception=throw_exception)

    if latest_version is None:
        return

    if current_version == latest_version:
        return

    running_under_virtualenv = getattr(sys, 'real_prefix', None) is not None

    if not running_under_virtualenv:
        logging.info('updating %s to version %s in user path', package, latest_version)

    return update_sdk(latest_version, user_path=not running_under_virtualenv, throw_exception=throw_exception)


def add_commands():
    from mali_commands import auth_commands, projects_commands, orgs_commands, experiments_commands, runcode_commands, \
        models_commands, data_commands

    cli.add_command(auth_commands)
    cli.add_command(projects_commands)
    cli.add_command(orgs_commands)
    cli.add_command(experiments_commands)
    cli.add_command(runcode_commands)
    cli.add_command(models_commands)
    cli.add_command(data_commands)


def main():
    self_update()
    add_commands()
    cli()


if __name__ == "__main__":
    main()



# coding=utf-8
from .mysql_ref_container import MySqlRefsContainer
from .spanner_ref_container import SpannerRefContainer
from .datastore_ref_container import DatastoreRefContainer
from .backend_ref_container import BackendRefContainer

try:
    from .gae_datastore_ref_container import GAEDatastoreRefContainer
except ImportError:
    GAEDatastoreRefContainer = None



# coding=utf-8
from ...connection_mixin import ConnectionMixin
from .tree_cache import TreeCache


class SqliteTreeCache(ConnectionMixin, TreeCache):
    create_tree_sql = """
      CREATE TABLE
      IF NOT EXISTS tree (
        name TEXT PRIMARY KEY,
        files BLOB,
        sha TEXT
      );
    """

    def __init__(self, connection):
        super(SqliteTreeCache, self).__init__(connection)
        self._create_table_if_needed()

    def _create_table_if_needed(self):
        with self._connection.get_cursor() as c:
            self._connection.execute_multi(c, self.create_tree_sql)
            self._connection.commit()

    def _set_files(self, trees_objects):
        import sqlite3

        inserts = []
        for path, tree in trees_objects.items():
            blob = '\n'.join(self.build_blob(tree))

            inserts.append((path.decode('utf8'), sqlite3.Binary(blob.encode('utf8')), tree.id.decode('utf8')))

        with self._connection.get_cursor() as c:
            c.executemany('INSERT OR REPLACE INTO [tree] (name, files, sha) VALUES (?, ?, ?)', inserts)

    @classmethod
    def decode_sqlite_to_str(cls, val):
        return str(val)

    def _get_files(self, scans):
        with self._connection.get_cursor() as c:
            params = ",".join(['?'] * len(scans))
            paths = [path.decode('utf8') for path, _, _ in scans]
            c.execute("select [name], [files] from tree where [name] in (%s)" % params, paths)
            files_data = c.fetchall()
            return {self.decode_sqlite_to_str(files_row['name']): self._files_sha_array_to_dict(self.decode_sqlite_to_str(files_row['files'])) for files_row in files_data}

    def get_commit_id(self):
        with self._connection.get_cursor() as c:
            c.execute("select [sha] from [tree] where [name] = ? limit 1", ('', ))
            files_data = c.fetchone()

            commit_id = files_data['sha']

            return commit_id.encode('utf8')



# coding=utf-8
from tests_data_management.base import BaseTest


class TestDataManagement(BaseTest):

    def testBasicFlow(self):
        """Test basic flow:
        - Create a data volume
        - Add a file
        - Add a metadata
        - Commit
        - Clone
        """
        print('Testing basic flow')

        self.populate_data([
            'foo.txt',
        ])

        volume_id = self.create_data_volume()
        self.add_file(volume_id, ['foo.txt'])
        self.add_metadata(volume_id, ['foo.txt'], {'name': 'foo'})
        commit_id = self.commit(volume_id)
        self.clone(volume_id, self.clone_dir, 'name:foo version:%s' % commit_id)

        self.assertEqualDir(self.data_dir, self.clone_dir)



from contextlib import closing
from mali_commands.legit.dulwich import repo
from tests.base import BaseTest
from mali_commands.data_volume import create_data_volume, with_repo
import tempfile
from mali_commands.legit.object_store.gcs.gcs_object_store import status_404_to_204, return_request_result
from mali_commands.legit.dulwich import porcelain
import fudge
from fudge import patched_context
import os


class TestData(BaseTest):
    @classmethod
    def get_repo(cls, temp_path=None, volume_id=None, linked=False, description=None, display_name=None):
        temp_path = temp_path or tempfile.tempdir
        volume_id = volume_id or cls.some_random_shit_number_int63()
        data_path = temp_path
        description = description or cls.some_random_shit(size=50)
        display_name = display_name or cls.some_random_shit(size=10)
        res = create_data_volume(volume_id, data_path, linked, display_name, description)
        return res

    def test_404_to_204(self):
        resp = self.get_obj(status=404)
        status_404_to_204(resp)
        self.assertEqual(resp.status, 204)

    def test_non_404_to_204(self):
        for status in [400, 300, 200]:
            resp = self.get_obj(status=status)
            status_404_to_204(resp)
            self.assertEqual(resp.status, status)

    def test_return_request_result(self):
        t1 = self.some_random_shit(size=30)
        t2 = self.some_random_shit(size=30)
        x1, x2 = return_request_result(t1, t2)
        self.assertEqual(x1, t1)
        self.assertEqual(x2, t2)

    def test_repo_add_objects(self):
        temp_path = tempfile.tempdir
        volume_id = self.some_random_shit_number_int63()
        data_path = temp_path
        linked = False
        description = self.some_random_shit(size=50)
        display_name = self.some_random_shit(size=10)
        res = create_data_volume(volume_id, data_path, linked, display_name, description)
        self.assertIsNotNone(res)
        with with_repo(None, volume_id) as rep:
            self.assertIsNotNone(rep)

    def test_porcelan_add_stage(self):
        import os
        import tempfile
        file_count = 10
        is_embedded = True
        volume_id = self.some_random_shit_number_int63()
        repo_config = self.get_repo(volume_id=volume_id, linked=is_embedded)

        def make_temp_files(count):
            for i in range(count):
                f = tempfile.NamedTemporaryFile(suffix='.txt', prefix='test_enumerate_paths__', delete=False)
                f.write(bytearray("Hello World!\n{}\n".format(i), 'utf8'))
                f.close()
                yield f.name

        tempfiles = list(make_temp_files(file_count))
        with with_repo(None, volume_id) as rep:
            rep.data_path = '/'

            def patched_stage(inst, paths, embedded):
                for path in paths:
                    self.assertIn('/' + path, tempfiles)
                self.assertEqual(len(paths), len(tempfiles))

            with patched_context(rep, 'stage', patched_stage):
                porcelain.add(rep, tempfiles, is_embedded)

    def test_porcelan_stage_generate_files(self):
        import tempfile
        file_count = 10
        is_embedded = True
        volume_id = self.some_random_shit_number_int63()
        repo_config = self.get_repo(volume_id=volume_id, linked=is_embedded)

        def make_temp_files(count):
            for i in range(count):
                f = tempfile.NamedTemporaryFile(suffix='.txt', prefix='test_enumerate_paths__', delete=False)
                f.write(bytearray("Hello World!\n{}\n".format(i), 'utf8'))
                f.close()
                yield f.name

        tempfiles = list(make_temp_files(file_count))
        with with_repo(None, volume_id) as rep1:
            rep1.data_path = '/'
            rep1.path = '/'

            def patched_stage1(inst, files):
                for file in files:
                    self.assertIn(file.full_path, tempfiles)
                    self.assertIn(os.path.join('/', file.tree_path.decode('utf-8')), tempfiles)
                self.assertEqual(len(files), len(tempfiles))

            def patched_set_entities(inst, index, en):
                for file in en:
                    self.assertIn('/' + file.decode('utf-8'), tempfiles)
                self.assertEqual(len(en.keys()), len(tempfiles))

            with patched_context(rep1, '_upload_staged_files', patched_stage1):
                with patched_context(rep1, '_set_entries', patched_set_entities):
                    porcelain.add(rep1, tempfiles, is_embedded)

    def test_stage_file_builder(self):
        import tempfile
        file_count = 10
        is_embedded = True
        volume_id = self.some_random_shit_number_int63()
        repo_config = self.get_repo(volume_id=volume_id, linked=is_embedded)

        def make_temp_files(count):
            for i in range(count):
                f = tempfile.NamedTemporaryFile(suffix='.txt', prefix='test_enumerate_paths__', delete=False)
                f.write(bytearray("Hello World!\n{}\n".format(i), 'utf8'))
                f.close()
                yield f.name

        tempfiles = list(make_temp_files(file_count))
        with with_repo(None, volume_id) as rep:
            rep.data_path = '/'
            rep.path = '/'
            index = rep.open_index()

            closure_dict = {}

            def patched_stage(inst, paths, embedded):
                for path in paths:
                    self.assertIn('/' + path, tempfiles)
                self.assertEqual(len(paths), len(tempfiles))
                closure_dict['rel_paths'] = paths

            with patched_context(rep, 'stage', patched_stage):
                porcelain.add(rep, tempfiles, is_embedded)

            relpaths = closure_dict['rel_paths']
            files = list(rep._build_stage_files(relpaths, index, is_embedded))
            self.assertEqual(len(files), len(tempfiles))



# coding=utf-8
from abc import ABCMeta, abstractmethod
from contextlib import closing
from ..connection_mixin import ConnectionMixin


class BaseMLIndex(ConnectionMixin):
    __metaclass__ = ABCMeta

    def __init__(self, connection):
        super(BaseMLIndex, self).__init__(connection)

        self._create_table_if_needed()

    def _get_hash_by_name_query(self):
        pass

    def get_hash_by_name(self, name):
        with closing(self._conn.cursor()) as c:
            c.execute(self._get_hash_by_name_query(), (name.decode('utf8'), ))

            result = c.fetchone()
            return result[0].encode('utf8') if result is not None else None

    @abstractmethod
    def _create_table_if_needed(self):
        """

        :return:
        """

    @abstractmethod
    def begin_commit(self, tree_id):
        """

        :return:
        """

    @abstractmethod
    def end_commit(self):
        """

        :return:
        """

    @classmethod
    def _decode_entries(cls, entries):
        values = []
        for name, entry in entries.items():
            ctime, mtime, dev, ino, mode, uid, gid, size, sha, flags, url = entry

            name = name.decode('utf8')
            sha = sha.decode('utf8')

            row = (name, sha, ctime, mtime, mode, uid, gid, size, url)

            values.append(row)

        return values



# coding=utf-8
import logging
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager


class BaseConnection(object):
    __metaclass__ = ABCMeta

    def __init__(self, data_volume_config, **kwargs):
        self.read_only = kwargs.get('read_only', False)
        self.data_volume_config = data_volume_config
        self.__conn = self._create_connection(**kwargs)
        self.__in_transactions = False
        self.__cursor = None

    @contextmanager
    def get_cursor(self):
        should_close = self.__cursor is None
        return_cursor = None
        try:
            return_cursor = self._create_cursor() if self.__cursor is None else self.__cursor
            yield return_cursor
        finally:
            if should_close and return_cursor is not None and hasattr(return_cursor, 'close'):
                return_cursor.close()

    def commit(self):
        if self.__in_transactions:
            logging.info('db in transaction commit will be skipped')
            return

        self._commit()

    def close(self):
        if self.__conn is not None and hasattr(self.__conn, 'close'):
            self.__conn.close()

    @property
    def _native_conn(self):
        return self.__conn

    @abstractmethod
    def _create_cursor(self):
        """

        :return:
        """

    @abstractmethod
    def _commit(self):
        """

        :return:
        """

    @abstractmethod
    def _rollback(self):
        """

        :return:
        """

    @abstractmethod
    def create_sql_helper(self):
        """

        :return:
        """

    @abstractmethod
    def _create_connection(self, **kwargs):
        """

        :param kwargs:
        :return:
        """

    def start_transactions(self):
        self.__in_transactions = True
        self.__cursor = self._create_cursor()

    def end_transactions(self):
        self._commit()
        self.__in_transactions = False

        if hasattr(self.__cursor, 'close'):
            self.__cursor.close()

    def rollback_transactions(self):
        self._rollback()
        self.__in_transactions = False
        if hasattr(self.__cursor, 'close'):
            self.__cursor.close()



# coding=utf-8
import json
import logging
import uuid
import os
from .connection_mixin import ConnectionMixin


class BigQueryOperationError(Exception):
    pass


class BigQueryMixin(ConnectionMixin):
    @classmethod
    def handle_bq_errors(cls, async_job_or_errors):
        if isinstance(async_job_or_errors, list):
            logging.debug('finish job: errors: %s', async_job_or_errors)
            errors = async_job_or_errors
        else:
            logging.debug('finish job:\n %s', async_job_or_errors.result())
            errors = async_job_or_errors.errors

        if not errors:
            return

        error_msg = json.dumps(errors)

        raise BigQueryOperationError(error_msg)

    def _query_async(self, query, query_parameters=(), use_legacy_sql=None):
        debug_params = [param.to_api_repr() for param in query_parameters]
        logging.info('query async: %s, params: %s, legacy_sql: %s', query, debug_params, use_legacy_sql)
        query_id = str(uuid.uuid4())

        bq_client = self._connection

        query_job = bq_client.run_async_query(
            query_id, query, query_parameters=query_parameters)

        if use_legacy_sql is not None:
            query_job.use_legacy_sql = use_legacy_sql

        return query_job

    @classmethod
    def __return_results_iter(cls, query_results, max_results, start_index, process_row):
        page_iter = query_results.fetch_data(max_results=max_results, start_index=start_index)

        total = 0

        for result in page_iter:
            total += 1

            if max_results is not None and total > max_results:
                break

            if not process_row:
                yield result
                continue

            yield process_row(result, query_results.schema)

    def _query_sync(
            self, query, query_parameters=(), use_legacy_sql=None,
            process_row=None, max_results=None, start_index=None):
        query_job = self._query_async(query, query_parameters, use_legacy_sql=use_legacy_sql)
        query_job.begin()
        query_results = query_job.query_results()

        self.handle_bq_errors(query_job)

        return self.__return_results_iter(query_results, max_results, start_index, process_row), query_results.total_rows

    def _copy_table(self, dest_table, src_table, write_disposition, create_disposition):
        logging.info(
            'copy %s => %s write_disposition: %s, create_disposition: %s',
            src_table.name, dest_table.name, write_disposition, create_disposition)

        copy_id = str(uuid.uuid4())

        bq_client = self._connection

        copy_job = bq_client.copy_table(copy_id, dest_table, src_table)
        copy_job.write_disposition = write_disposition
        copy_job.create_disposition = create_disposition

        return copy_job

    def _sync_copy_table(self, dest_table, src_table, write_disposition, create_disposition):
        copy_job = self._copy_table(dest_table, src_table, write_disposition, create_disposition)
        copy_job.begin()

        self.handle_bq_errors(copy_job.result())

    @classmethod
    def valid_table_name_random_token(cls, size=8):
        from base58 import b58encode

        return b58encode(os.urandom(size))

    @classmethod
    def _get_table_name(cls, prefix, name):
        table_full_name = '{prefix}_{name}'.format(prefix=prefix, name=name)

        return table_full_name

    def _get_table(self, name, schema=()):
        table_full_name = self._get_table_name(self._connection.table_prefix, name)

        with self._connection.get_cursor() as bq_dataset:
            return bq_dataset.table(table_full_name, schema=schema)

    def _create_specific_table(self, name, schema):
        import google.api.core.exceptions

        table = self._get_table(name)
        table.schema = schema

        try:
            table.create()
        except google.api.core.exceptions.Conflict:
            table.reload()

        return table

    def _copy_table_data(self, src_query, src_query_params, dest_table, write_disposition=None):
        from google.cloud.bigquery.job import WriteDisposition

        write_disposition = write_disposition or WriteDisposition.WRITE_APPEND

        logging.info(
            'copy_table_data to %s filter query: %s params: %s (%s)',
            dest_table.name, src_query, src_query_params, write_disposition)

        query_job = self._query_async(src_query, query_parameters=src_query_params)
        query_job.destination = dest_table
        query_job.write_disposition = write_disposition
        query_job.begin()
        self.handle_bq_errors(query_job.result())

    @classmethod
    def bq_field_name_to_common_name(cls, name):
        return '@' + name[1:] if name in ['_commit_sha', '_sha', '_hash', '_url'] else name

    @classmethod
    def common_name_to_bq_field_name(cls, name):
        return '_' + name[1:] if name.startswith('@') else name

    @classmethod
    def build_dict(cls, row, schema):
        return {
            cls.bq_field_name_to_common_name(field.name): val for field, val in zip(schema, row)
        }



# coding=utf-8
from google.appengine.ext import ndb
from .connection_mixin import ConnectionMixin


class GAEDatastoreMixin(ConnectionMixin):
    @classmethod
    def build_key_from_volume_id(cls, org_id, volume_id, kind, name):
        parent_key = ndb.Key('DataVolume', volume_id, namespace=org_id)

        name = name or 1

        return ndb.Key(kind, name, parent=parent_key, namespace=org_id)

    def _build_key(self, name):
        return self.build_key_from_volume_id(
            self._connection.data_volume_config.org,
            self._connection.data_volume_config.volume_id, self._entity_kind, name)

    @property
    def _entity_kind(self):
        raise NotImplementedError(self._entity_kind)
